
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <sys/wait.h>                       /* waitpid() sa */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>                          /* printf() */
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "initng_signal.h"
#include "initng_handler.h"
#include "initng_kill_handler.h"
#include "initng_global.h"
#include "initng_main.h"
#include "main.h"

static void sigchild(int sig);
static void sigwinch(int sig);
static void sigint(int sig);
static void sigalarm(int sig);

/*
 * These are gloval variables, that got set,
 * by the signals rescived.
 */
volatile int initng_signal_got_keyboard_request = FALSE;
volatile int initng_signal_got_ctrl_alt_del = FALSE;
volatile int initng_signal_got_alarm = FALSE;
volatile int initng_signal_got_sigchld = FALSE;

/*
 * If we got an chiled that died, this haldler is called.
 */
void initng_signal_handle_sigchild(void)
{
    int status;                 /* data got from waitpid, never used */
    pid_t killed;               /* pid of killed app */

    for (;;)
    {
        /* slaying zombies */
        do
        {
            killed = waitpid(-1, &status, WNOHANG);
        }
        while ((killed < 0) && (errno == EINTR));

        /* Nothing killed */
        if (killed == 0)
        {
            return;
        }

        /* unknown child */
        if ((killed < 0) && (errno == ECHILD))
        {
            return;
        }

        /* Error */
        if (killed < 0)
        {
            W_("Error, waitpid returned pid %d (%s)\n", killed,
               strerror(errno));
            return;
        }

        D_("sigchild(): PID %i KILLED WITH:\n"
           "WEXITSTATUS %i\n"
           "WIFEXITED %i\n"
           "WIFSIGNALED %i\n"
           "WTERMSIG %i\n"
           "WCOREDUMP %i\n"
           "WIFSTOPPED %i\n"
           "WSTOPSIG %i\n"
           "\n",
           killed,
           WEXITSTATUS(status),
           WIFEXITED(status),
           WIFSIGNALED(status),
           WTERMSIG(status),
           WCOREDUMP(status), WIFSTOPPED(status), WSTOPSIG(status));

        /*
         * call handle_killed_by_pid(), and will walk the active_db
         * setting the a_status of the service touched
         */
        initng_kill_handler_killed_by_pid(killed, WEXITSTATUS(status));
    }
}

/* called by signal SIGCHILD */
static void sigsegv(int sig)
{
    (void) sig;
    printf("SEGFAULTED!\n");
    initng_main_segfault();
}


/* called by signal SIGCHILD */
static void sigchild(int sig)
{
    (void) sig;
    initng_global_set_interrupt();
    D_("sigchld():     Got Sigchld\n");
    initng_signal_got_sigchld = TRUE;
}


/* called by signal SIGWINCH */
static void sigwinch(int sig)
{
    (void) sig;
    initng_global_set_interrupt();
    D_("sigwinch():     Got Keyb Request\n");
    initng_signal_got_keyboard_request = TRUE;
}

/* sig alarm got */
static void sigalarm(int sig)
{
    (void) sig;
    initng_global_set_interrupt();
    D_("sigalrm(): Got alarm\n");
    initng_signal_got_alarm = TRUE;
}

/* called by signal SIGINT */
static void sigint(int sig)
{
    (void) sig;
    initng_global_set_interrupt();
    D_("sigint():     Got CTRL_ALT_DELETE\n");
    initng_signal_got_ctrl_alt_del = TRUE;
}

/*
 * This is called in main() initziation, and will enable signals when it happends.
 */
void initng_signal_enable(void)
{
    D_("enable_signals();\n");

    /* clear interrupt cache */
    /* TODO FIGUREITOUT - Catch signals */
    /*  signal(SIGPWR,sighandler); don't know what to do about it */
    /*  signal(SIGHUP,sighandler); ??? */
    /* SA_NOCLDSTOP -->  Get notification when childs stops living */
    /* SA_RESTART --> make certain system calls restartable across signals */
    /*   Normally if a program is in a system call and a signal is */
    /*   delivered, the system call exits with and error sets errno */
    /*   to EINTR. This is quite annoying in practice, as *every* */
    /*   system call has to done in a loop, retrying if the error */
    /*   EINTR. By setting SA_RESTART, system calls will "restart", */
    /*   continuing normally when your signal handler returns. */

    /* create a signal set */
    sigemptyset(&sa.sa_mask);

    /* i don't get this one */
    sa.sa_sigaction = 0;


    /* SA_NOCLDSTOP = Don't give initng signal if we kill the app with SIGSTOP */
    /* SA_RESTART = call signal over again next time */
    sa.sa_flags = SA_RESTART | SA_NOCLDSTOP;

    if (g.i_am_init)
    {
        sa.sa_handler = sigsegv;
        sigaction(SIGSEGV, &sa, 0);
        sigaction(SIGABRT, &sa, 0);
    }

    sa.sa_handler = sigchild;
    sigaction(SIGCHLD, &sa, 0);             /* Dead children */

    sa.sa_handler = sigint;
    sigaction(SIGINT, &sa, 0);              /* ctrl-alt-del */

    sa.sa_handler = sigwinch;
    sigaction(SIGWINCH, &sa, 0);            /* keyboard request */

    sa.sa_handler = sigalarm;
    sigaction(SIGALRM, &sa, 0);             /* alarm, something has to be checked */

}

/* 
 * Disable signals, called when initng are exiting.
 * somethimes, we don't want to be disturbed with signals.
 */
void initng_signal_disable(void)
{
    D_("disable_signals();\n");

    /* create a signal set */
    sigemptyset(&sa.sa_mask);

    sa.sa_flags = SA_RESTART | SA_NOCLDSTOP;

    /* i don't get this one */
    sa.sa_sigaction = 0;
    sa.sa_handler = SIG_DFL;
    sigaction(SIGSEGV, &sa, 0);
    sigaction(SIGABRT, &sa, 0);
    sigaction(SIGCHLD, &sa, 0);             /* Dead children */
    sigaction(SIGINT, &sa, 0);              /* ctrl-alt-del */
    sigaction(SIGWINCH, &sa, 0);            /* keyboard request */
    sigaction(SIGALRM, &sa, 0);             /* alarm, something has to be checked */

}
