
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_global.h"
#include "initng_static_process_types.h"
#include "initng_process_db.h"
#include "initng_common.h"
#include "initng_static_states.h"
#include "initng_fd.h"
#include "initng_static_data_id.h"

static void handle_killed_daemon(active_db_h * killed_service,
                                 process_h * process);
static void handle_killed_start(active_db_h * killed_service,
                                process_h * process);
static void handle_killed_stop(active_db_h * killed_service,
                               process_h * process);


ptype_h T_START = { "start", &handle_killed_start };
ptype_h T_STOP = { "stop", &handle_killed_stop };
ptype_h T_DAEMON = { "daemon", &handle_killed_daemon };
ptype_h T_KILL = { "kill", NULL };

void initng_static_process_types_add_default(void)
{
    initng_process_db_ptype_add(&T_START);
    initng_process_db_ptype_add(&T_STOP);
    initng_process_db_ptype_add(&T_DAEMON);
    initng_process_db_ptype_add(&T_KILL);
}
static void handle_killed_daemon(active_db_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(service->current_state);
    assert(service->current_state->state_name);
    assert(process);

    D_("(%s): initial status: \"%s\".\n",
       service->name, service->current_state->state_name);

    /* Set the universal variable, that signalize that something happend */
    initng_global_set_interrupt();

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
    {
        F_(" daemon %s, Returned with exit %i.\n", service->name,
           process->r_code);

	initng_common_mark_service(service, &DAEMON_EXIT_BAD);
	list_del(&process->list);
	initng_process_db_free(process);
	return;
    }

    /* free the process, or start_service will fail */
    list_del(&process->list);
    initng_process_db_free(process);

    /* OK, now accept this service as STOPPED, set this status */
    initng_common_mark_service(service, &STOPPED);
}

/*
 * handle_killed_start(), description:
 *
 * A process dies, and kill_handler walks the active_db looking for a match
 * on that pid, if it founds a match for a active->start_process, this
 * function is called, with a pointer to that active_service to
 * handle the situation.
 */
static void handle_killed_start(active_db_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(service->current_state);
    assert(service->current_state->state_name);
    assert(process);

    D_("handle_killed_start(%s): initial status: \"%s\".\n",
       service->name, service->current_state->state_name);

    /* Set the universal variable, that signalize that something happend */
    initng_global_set_interrupt();

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
        F_(" start %s, Returned with exit %i.\n", service->name,
           process->r_code);

    /*
     * If service is stopping, ignore this signal
     */
    if (service->current_state != &STARTING)
    {
        F_("Start exited!, and service is not marked starting!\n");
        return;
    }


    /*
     * Make sure r_code don't signal error (can be override by UP_ON_FAILURE.
     */
    if (process->r_code && !initng_active_db_is(&UP_ON_FAILURE, service))
    {

        initng_common_mark_service(service, &FAIL_STARTING);
        list_del(&process->list);
        initng_process_db_free(process);
        return;
    }

    /* OK! now service is DONE! */
    initng_common_mark_service(service, &DONE);

    /* free the process struct, to spare memory */
    list_del(&process->list);
    initng_process_db_free(process);
}


/* to do when a stop action dies */
static void handle_killed_stop(active_db_h * service, process_h * process)
{
    assert(service);
    assert(service->name);
    assert(service->current_state);
    assert(service->current_state->state_name);
    assert(process);

    D_("(%s);\n", service->name);

    if (service->current_state != &STOPPING)
    {
        F_("stop service died, but service is not status STOPPING!\n");
    }

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (process->r_code > 0)
        F_(" stop %s, Returned with exit %i.\n", service->name,
           process->r_code);


    /* free stopped process data */
    list_del(&process->list);
    initng_process_db_free(process);

    /* mark service stopped */
    initng_common_mark_service(service, &STOPPED);
}
